REBOL [
	Title: "REBOL Mezzanine Functions: Help"
	Rights: "Copyright REBOL Technologies 1997-2002"
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

dump-obj: func [
	"Returns a block of information about an object."
	obj [object!] 
	/match "Include only those that match a string or datatype" pat
	/local clip-str form-val form-pad words vals str wild
][
	clip-str: func [str] [
		; Keep string to one line.
		trim/lines str 
		if (length? str) > 50 [str: append copy/part str 50 "..."]
		str
	]

	form-val: func [val] [
		; Form a limited string from the value provided.
		if any-block? :val [return reform ["length:" length? val]]
		if image? :val [return reform ["size:" val/size]]
		if any-function? :val [
			val: third :val
			if block? val/1 [val: next val]
			return clip-str either string? val/1 [copy val/1][mold val]
		]
		if object? :val [val: next first val]
		clip-str mold :val
	]

	form-pad: func [val size] [
		; Form a value with fixed size (space padding follows).
		val: form val
		insert/dup tail val #" " size - length? val
		val
	]

	words: first obj
	vals: next second obj
	obj: copy []
	wild: all [string? pat  find pat "*"]
	foreach word next words [
		type: type?/word pick vals 1
		str: form word
		if any [
			not match
			all [
				not unset? pick vals 1
				either string? :pat [
					either wild [
						tail? any [find/any/match str pat pat]
					][
						find str pat
					]
				][
					all [
						datatype? get :pat
						type = :pat
					]
				]
			]
		][
			str: form-pad word 15
			append str #" "
			append str form-pad type 10 - ((length? str) - 15)
			append obj reform [
				"  " str
				if type <> 'unset! [form-val pick vals 1]
				newline
			]
		]
		vals: next vals
	]
	obj
]

?: help: func [
	"Prints information about words and values."
	'word [any-type!]
	/local value args item type-name refmode types attrs rtype
][
	if unset? get/any 'word [
		print trim/auto {
			To use HELP, supply a word or value as its
			argument:
			
				help insert
				help system
				help system/script

			To view all words that match a pattern use a
			string or partial word:

				help "path"
				help to-

			To see words with values of a specific datatype:

				help native!
				help datatype!

			Word completion:

				The command line can perform word
				completion. Type a few chars and press TAB
				to complete the word. If nothing happens,
				there may be more than one word that
				matches. Press TAB again to see choices.

				Local filenames can also be completed.
				Begin the filename with a %.

			Other useful functions:

				about - see general product info
				usage - view program options
				license - show terms of user license
				source func - view source of a function
				upgrade - updates your copy of REBOL
			
			More information: http://www.rebol.com/docs.html
		}
		exit
	]

	; If arg is an undefined word, just make it into a string:
	if all [word? :word not value? :word] [word: mold :word]

	; If arg is a string or datatype! word, search the system:
	if any [string? :word all [word? :word datatype? get :word]] [
;			types: copy []
;			attrs: second system/words
;			foreach item first system/words [
;				value: copy "              "
;				change value :item
;				if all [not unset? first attrs
;					any [
;						all [string? :word find value word]
;						all [not string? :word datatype? get :word (get :word) = type? first attrs]
;					]
;				][
;					repend value [" (" type? first attrs ")"]
;					append types value
;				]
;				attrs: next attrs
;			]
		types: dump-obj/match system/words :word
		sort types
		if not empty? types [
			print ["Found these words:" newline types]
			exit
		]
		print ["No information on" word "(word has no value)"]
		exit
	]

	; Print type name with proper singular article:
	type-name: func [value] [
		value: mold type? :value
		clear back tail value
		join either find "aeiou" first value ["an "]["a "] value
	]

	; Print literal values:
	if not any [word? :word path? :word][
		print [mold :word "is" type-name :word]
		exit
	]

	; Get value (may be a function, so handle with ":")
	value: either path? :word [first reduce reduce [word]][get :word]
	if not any-function? :value [
		prin [uppercase mold word "is" type-name :value "of value: "]
		print either object? value [print "" dump-obj value][mold :value]
		exit
	]       

	;-- Print info about function:
	;print [word mold first :value]
	args: third :value
	prin "USAGE:^/^-"
	if not op? :value [prin append uppercase mold word " "]
	while [not tail? args] [
		item: first args
		if :item = /local [break]
		if any [all [any-word? :item  not set-word? :item] refinement? :item] [
			prin append mold :item " " ; bug in prin?!!!
			if op? :value [prin append uppercase mold word " " value: none]
		]
		args: next args
	]
	print ""
	args: head args
	value: get word

	print "^/DESCRIPTION:"
	either string? pick args 1 [
		print [tab first args]
		args: next args
	][
		print "^-(undocumented)"
	]
	print [tab uppercase mold word "is" type-name :value "value."]

	if block? pick args 1 [  ; if attributes, save for latter
		attrs: first args
		args: next args
	]
	if tail? args [exit]

	while [not tail? args] [
		item: first args
		args: next args
		if :item = /local [break]
		either not refinement? :item [
			all [set-word? :item :item = to-set-word 'return block? first args rtype: first args]
			if none? refmode [
				print "^/ARGUMENTS:"
				refmode: 'args
			]
		][
			if refmode <> 'refs [
				print "^/REFINEMENTS:"
				refmode: 'refs
			]

		]
		either refinement? :item [
			prin [tab mold item]
			if string? pick args 1 [prin [" --" first args]  args: next args]
			print ""
		][
			if all [any-word? :item not set-word? :item] [
				if refmode = 'refs [prin tab]
				prin [tab :item "-- "]
				types: if block? pick args 1 [args: next args first back args]
				if string? pick args 1 [prin [first args ""] args: next args]
				if not types [types: 'any]
				prin rejoin ["(Type: " types ")"]
				print ""
			]
		]
	]
	if rtype [print ["^/RETURNS:^/^-" rtype]]

	if attrs [
		print "^/(SPECIAL ATTRIBUTES)"
		while [not tail? attrs] [
			value: first attrs
			attrs: next attrs
			if any-word? value [
				prin [tab value]
				if string? pick attrs 1 [
					prin [" -- " first attrs]
					attrs: next attrs
				]
				print ""
			]
		]
	]
	exit
]

about: func [
	"Information about REBOL"
][
	print join "REBOL/" [system/product #" " system/version #" " system/build/date]
	print join "Copyright 2000-" [system/build/year " REBOL Technologies.  All rights reserved."]
	print "REBOL is a trademark of REBOL Technologies. WWW.REBOL.COM"
	;if view? [print "^/Type DESKTOP or SET-USER for settings."]
]

usage: func [
	"Prints command-line arguments."
][
	print trim/auto {
	The command line usage is:

		REBOL <options> <script> <arguments>

	All fields are optional. Supported options are:

		--cgi (-c)       Check for CGI input
		--do expr        Evaluate expression
		--link url       Connect to Link
		--help (-?)      Display this usage information
		--nowindow (-w)  Do not open a window
		--noinstall (-i) Do not install (Link, View)
		--quiet (-q)     Don't print banners
		--reinstall (+i) Force an install (Link, View)
		--script file    Explicitly specify script
		--secure level   Set security: allow ask throw quit
		--trace (-t)     Enable trace mode
		--uninstall (-u) Uninstall REBOL (Link, View)
		--version tuple  Minimum version of script, when URL (View)
		--noviewtop (-v) Do not start viewtop (view desktop)

	Special command line options:

		+q               Force not quiet (Link, View)
		-s               No security
		+s               Full security
		-- args          Provide args without a script

	Examples:

		REBOL script.r
		REBOL -s script.r
		REBOL script.r 10:30 test@domain.dom
		REBOL --do "verbose: true" script.r
		REBOL -cswq
		REBOL --cgi --secure throw --script cgi.r "debug: true"
		REBOL --version 1.2.3 http://www.rebol.net/test.r ; view only
	}
]

license: func [
	"Prints the REBOL/core license agreement."
][
	print system/license
]

upgrade: func [
	"Download a new version of REBOL if available."
	/local platforms fields me message
][
	do http://www.rebol.com/releases.html
]

source: func [
	"Prints the source code for a word."
	'word [word!]
][
	prin join word ": "
	if not value? word [print "undefined" exit]
	either any [native? get word op? get word action? get word][
		print ["native" mold third get word]
	][print mold get word]
]

what: func [
	{Prints a list of globally-defined functions.}
	/local vals args here total
][
	total: copy []
	vals: second system/words
	foreach word first system/words [
		if any-function? first vals [
			args: first first vals
			if here: find args /local [args: copy/part args here]
			append total reduce [word mold args]
		]
		vals: next vals
	]
	foreach [word args] sort/skip total 2 [print [word args]]
	exit
]

